#include <list>

#include "mpi_collective.hpp"

// bb aux alloc, free & reset
void pass_mpi_collective::alloc_bb_aux(function *fun)
{
	basic_block bb;

	FOR_ALL_BB_FN(bb, fun)
	{
		bb->aux = (bb_data *) new bb_data();
		bb_data *data = ((bb_data *) bb->aux);
		data->mpi_code = LAST_AND_UNUSED_MPI_COLLECTIVE_CODE;
		bitmap_initialize(&data->dom, &bitmap_default_obstack);
		bitmap_initialize(&data->post_dom, &bitmap_default_obstack);
		bitmap_initialize(&data->dom_front, &bitmap_default_obstack);
		bitmap_initialize(&data->post_dom_front, &bitmap_default_obstack);
		for (int i = 0; i < LAST_AND_UNUSED_MPI_COLLECTIVE_CODE; ++i)
		{
			data->collective_rank[i] = 0;
		}
		bitmap_initialize(&data->seens, &bitmap_default_obstack);
	}

	reset_bb_mark(fun);
}

void pass_mpi_collective::free_bb_aux(function *fun)
{
	basic_block bb;
	FOR_ALL_BB_FN(bb, fun)
	{
		if (bb->aux != NULL)
		{
			bb_data *data = ((bb_data *) bb->aux);
			bitmap_clear(&data->dom);
			bitmap_clear(&data->post_dom);
			bitmap_clear(&data->dom_front);
			bitmap_clear(&data->post_dom_front);
			bitmap_clear(&data->seens);
			delete (bb_data *) bb->aux;
			bb->aux = NULL;
		}
	}
}

void pass_mpi_collective::reset_bb_mark(function *fun)
{
	basic_block bb;

	FOR_ALL_BB_FN(bb, fun)
	{
		((bb_data *) bb->aux)->mark1 = 0;
		((bb_data *) bb->aux)->mark2 = 0;
	}
}

// edge aux alloc & free
void pass_mpi_collective::alloc_edge_aux(function *fun)
{
	basic_block bb;
	edge e;
	edge_iterator ei;

	FOR_ALL_BB_FN(bb, fun)
	{
		FOR_EACH_EDGE(e, ei, bb->succs)
		{
			e->aux = new edge_data();
		}
	}
}

void pass_mpi_collective::free_edge_aux(function *fun)
{
	basic_block bb;
	edge e;
	edge_iterator ei;

	FOR_ALL_BB_FN(bb, fun)
	{
		FOR_EACH_EDGE(e, ei, bb->succs)
		{
			if (e->aux != NULL)
			{
				delete (edge_data *) e->aux;
				e->aux = NULL;
			}
		}
	}
}

// loop detection
// naive version of mark_edge (work but Omax(2^n))
void pass_mpi_collective::__mark_edge_naif(function *fun)
{
	__mark_edge_naif(ENTRY_BLOCK_PTR_FOR_FN(fun));
}

void pass_mpi_collective::__mark_edge_naif(basic_block bb)
{
	edge e;
	edge_iterator ei;

	((bb_data *) bb->aux)->mark1 = 1;
	FOR_EACH_EDGE(e, ei, bb->succs)
	{
		if (((bb_data *) e->dest->aux)->mark1)
		{
			((edge_data *) e->aux)->loop = true;
		}
		else
		{
			__mark_edge_naif(e->dest);
		}
	}
	((bb_data *) bb->aux)->mark1 = 0;
}

void pass_mpi_collective::__mark_edge_test(function *fun)
{

	edge e1;
	edge_iterator ei1;
	std::list<basic_block> nodes1;
	basic_block bb;

	edge e2;
	edge_iterator ei2;
	std::list<basic_block> nodes2;
	basic_block cursor;

	int index;

	index = 0;

	nodes1.push_back(ENTRY_BLOCK_PTR_FOR_FN(fun));
	((bb_data *) ENTRY_BLOCK_PTR_FOR_FN(fun)->aux)->mark1 = 1;

	while (!nodes1.empty())
	{
		bb = nodes1.front();
		nodes1.pop_front();

		FOR_EACH_EDGE(e1, ei1, bb->succs)
		{
			if (((bb_data *) e1->dest->aux)->mark1)
			{
				continue;
			}
			((bb_data *) bb->aux)->mark1 = 1;
			nodes1.push_back(e1->dest);
		}

		if (EDGE_COUNT(bb->preds) < 2)
		{
			continue;
		}

		++index;
		nodes2.clear();
		nodes2.push_back(bb);
		((bb_data *) bb->aux)->mark2 = index;

		while (!nodes2.empty())
		{
			cursor = nodes2.front();
			nodes2.pop_front();

			FOR_EACH_EDGE(e2, ei2, cursor->succs)
			{
				if (e2->dest == bb)
				{
					((edge_data *) e2->aux)->loop = true;
				}
				if (((edge_data *) e2->aux)->loop)
				{
					continue;
				}

				if (((bb_data *) e2->dest->aux)->mark2 >= index)
				{
					continue;
				}
				((bb_data *) cursor->aux)->mark2 = index;
				nodes2.push_back(e2->dest);
			}
		}
	}
	reset_bb_mark(fun);
}

void pass_mpi_collective::mark_edge(function *fun)
{
	mark_edge(ENTRY_BLOCK_PTR_FOR_FN(fun));
	reset_bb_mark(fun);
}

void pass_mpi_collective::mark_edge(basic_block bb)
{
	edge e;
	edge_iterator ei;

	((bb_data *) bb->aux)->mark1 = 1;
	bitmap_set_bit(&((bb_data *) bb->aux)->seens, bb->index);
	FOR_EACH_EDGE(e, ei, bb->succs)
	{
		if (((edge_data *) e->aux)->loop)
		{
			continue;
		}
		if (bitmap_bit_p(&((bb_data *) bb->aux)->seens, e->dest->index))
		{
			((edge_data *) e->aux)->loop = true;
		}
		if (! ((bb_data *) e->dest->aux)->mark1)
		{
			bitmap_ior_into(&((bb_data *) e->dest->aux)->seens,
			                &((bb_data *) bb->aux)->seens);
			mark_edge(e->dest);
		}
	}
}
