// std::max
#include <algorithm>

#include "mpi_collective.hpp"

void pass_mpi_collective::rank_collective(function *fun)
{
	__rank_collective(EXIT_BLOCK_PTR_FOR_FN(fun));
	reset_bb_mark(fun);
}

void pass_mpi_collective::__rank_collective(basic_block bb)
{
	edge e;
	edge_iterator ei;

	((bb_data *) bb->aux)->mark1 = 1;
	FOR_EACH_EDGE(e, ei, bb->preds)
	{
		if (((edge_data *) e->aux)->loop)
		{
			continue;
		}

		if (! ((bb_data *) e->src->aux)->mark1)
		{
			__rank_collective(e->src);
		}

		for (int mpi_code = 0;
		     mpi_code < LAST_AND_UNUSED_MPI_COLLECTIVE_CODE;
		     ++mpi_code)
		{
			((bb_data *) bb->aux)->collective_rank[mpi_code] =
			    std::max(((bb_data *) bb->aux)->collective_rank[mpi_code],
			             ((bb_data *) e->src->aux)->collective_rank[mpi_code]);
			if (((bb_data *) bb->aux)->mpi_code == mpi_code)
			{
				((bb_data *) bb->aux)->collective_rank[mpi_code]++;
				collective_max_rank[mpi_code] =
				    std::max(collective_max_rank[mpi_code],
				             ((bb_data *) bb->aux)->collective_rank[mpi_code]);
				//printf("collective max rank for %i up to %i\n",
				//       mpi_code, collective_max_rank[mpi_code]);
			}
		}
	}
}

void pass_mpi_collective::better_rank_collective(function *fun)
{
	size_t i;
	int next_rank = 0;

	for (i = 0; i < LAST_AND_UNUSED_MPI_COLLECTIVE_CODE; ++i)
	{
		next_rank = __better_rank_collective(EXIT_BLOCK_PTR_FOR_FN(fun),
		                                     (enum mpi_collective_code) i,
		                                     next_rank);
	}
	reset_bb_mark(fun);
}

int pass_mpi_collective::__better_rank_collective(basic_block bb,
    mpi_collective_code mpi_code, int rank)
{
	edge e;
	edge_iterator ei;
	int next_rank;

	next_rank = rank;
	((bb_data *) bb->aux)->mark1 = mpi_code + 1;

	FOR_EACH_EDGE(e, ei, bb->preds)
	{
		if (((edge_data *) e->aux)->loop)
		{
			continue;
		}

		// cache or travell
		if (((bb_data *) e->src->aux)->mark1 <= mpi_code)
		{
			next_rank =
			    std::max(next_rank,
			             __better_rank_collective(e->src, mpi_code, rank));
		}
		else
		{
			next_rank =
			    std::max(next_rank,
			             ((bb_data *) e->src->aux)->collective_rank[mpi_code]);
		}
	}
	// set mpi code
	if (((bb_data *) bb->aux)->mpi_code == mpi_code)
	{
		next_rank++;
	}
	((bb_data *) bb->aux)->collective_rank[mpi_code] = next_rank;
	return next_rank;
}

void pass_mpi_collective::get_mpi_coll_rank(function *fun,
    int rank, int mpi_code, bitmap mpi_coll)
{
	basic_block bb;

	FOR_ALL_BB_FN(bb, fun)
	{
		if (((bb_data *) bb->aux)->mpi_code == mpi_code
		    && ((bb_data *) bb->aux)->collective_rank[mpi_code] == rank)
		{
			//printf("bb %i is coll %i of rank %i\n",
			//	bb->index, mpi_code, rank);
			bitmap_set_bit(mpi_coll, bb->index);
		}
	}
}

void pass_mpi_collective::raise_warning_if_mpi_wrong(function *fun)
{
	for (int mpi_code = 0; mpi_code < LAST_AND_UNUSED_MPI_COLLECTIVE_CODE;
	     ++mpi_code)
	{
		bitmap_head pds_head;
		bitmap_head pdf_head;
		bitmap_head nodes_head;

		bitmap pds = &pds_head;
		bitmap pdf = &pdf_head;
		bitmap nodes = &nodes_head;
		bitmap tmp = NULL;

		bitmap_initialize(pds, &bitmap_default_obstack);
		bitmap_initialize(pdf, &bitmap_default_obstack);
		bitmap_initialize(nodes, &bitmap_default_obstack);
		int stop = 0;

		for (int i = 1 ; i < collective_max_rank[mpi_code] + 1 ; i++)
		{
			printf("mpi_code: %i at rank: %i\n", mpi_code, i);
			// DONE: split in 2, get mpi set & find set post dom
			get_mpi_coll_rank(fun, i, mpi_code, nodes);
			get_post_dom_bitmap(fun, nodes, pds);
			get_post_dom_frontier(fun, pds, pdf);

			printf("----\n");
			debug_bitmap(nodes);
			debug_bitmap(pds);
			debug_bitmap(pdf);

			while (! bitmap_empty_p(pdf))
			{
				stop++;
				if (stop > 10)
				{ return; }

				printf("WARNING!!!!!!!!!!!!!!!!!!!!!!!!!!!");
				// print warning here

				bitmap_clear(nodes);
				bitmap_clear(pds);

				// swap nodes & pdf
				{
					tmp = nodes;
					nodes = pdf;
					pdf = tmp;
				}
				get_post_dom_bitmap(fun, nodes, pds);
				get_post_dom_frontier(fun, pds, pdf);

				printf("----\n");
				debug_bitmap(nodes);
				debug_bitmap(pds);
				debug_bitmap(pdf);
			}
			bitmap_clear(nodes);
			bitmap_clear(pds);
			bitmap_clear(pdf);
		}
	}
}

